(* $Id: godi_perform.ml 543 2005-04-04 17:15:16Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_query
open Printf

module StringSet = Set.Make(String) ;;

module StrMap = Map.Make(String) ;;


let default_protocol =
  lazy (
    `Forward (fun s -> prerr_string s; flush stderr)
  )
;;



(**********************************************************************)
(* Installed packages                                                 *)
(**********************************************************************)

let create_backup_name pkg_name s =
  let d = Filename.dirname s in
  let b = Filename.basename s in
  let b' = "." ^ b ^ "." ^ pkg_name ^ ".backup" in
  Filename.concat d b'
;;

let ph_pat_re = Str.regexp "%[FDBf]" ;;

let rewrite_placeholders cwd last_file pattern  =
  (* Pass last_file = "" if not available. *)
  Str.global_substitute
    ph_pat_re
    (fun s ->
       let m = Str.matched_string s in
       if m <> "%D" && last_file = "" then
	 user_error("Error in packing list - " ^ m ^
		    " cannot refer to last file");
       match m with
	   "%F" -> last_file
	 | "%D" -> cwd
	 | "%B" -> Filename.dirname(Filename.concat cwd last_file)
	 | "%f" -> Filename.basename(Filename.concat cwd last_file)
	 | _    -> assert false
    )
    pattern
;;

let is_prefix_of s t =
  String.length s <= String.length t &&
  String.sub t 0 (String.length s) = s
;;


type rm_mode = [ `Force | `Normal | `Simulate ] ;;

exception Soft_rm_error


let remove_pkg_files ?(mode = `Normal)
                     ?interactivity
		     ?(protocol = Lazy.force default_protocol) pkg =
  (* Deletes the files of the package, runs @unexec directives, restores
   * preserved files. Package scripts are not run.
   * "Soft" errors are printed, but mostly ignored.
   * The return value is [true] if no errors occured, [false] otherwise.
   * "Hard" errors trigger exceptions, however.
   *)
  let message = Godi_exec.collect protocol in
  let ipkg = installed_pkg pkg in
  let localbase = Godi_query.localbase() in
  let prefix = ref "" in
  let last_file = ref "" in
  let preserve = ref false in
  let plist = 
    Godi_plist.apply_ignore_directive (ipkg # inst_plist) in

  let dirs = ref StringSet.empty in
  let success = ref true in

  let check_prefix() =
    (* A missing or bad prefix is a fatal error, because we cannot continue! *)
    (* Missing prefix: *)
    if !prefix = "" then
      user_error "Error in packing list - no prefix set";
    (* Prefix must be a subdirectory of [LOCALBASE] *)
    if not (localbase = !prefix || is_prefix_of (localbase ^ "/") !prefix) then
      user_error "Error in packing list - prefix outside of managed directories";
  in

  (* The package files are deleted by iterating over the packing list,
   * and executing the directives.
   *)

  let csm_explained = ref false in

  List.iter
    (function
       | `Cwd d ->
	   if d <> "." then prefix := d
	     (* `Cwd "." is added by godi_create just before the "+" files *)
       | `Option `Preserve ->
	   preserve := true
	   (* Enables the check for preserved files *)
       | `File(name,checksum) ->
	   check_prefix();
	   last_file := name;
	   let fullname =
	     if Filename.is_relative name then
	       Filename.concat !prefix name
	     else
	       name in
	   (* Wrong PLISTs are handled as soft errors. It is tried to delete
	    * as much as possible. A message is printed, and [success] is
	    * reset.
	    *)
	   let can_remove1 =
	     not(Godi_file.file_is_dir fullname) || (
	       message("Error: File is a directory: " ^ fullname ^ "\n");
	       success := false;
	       false
	     ) in
	   let can_remove2 =
	     can_remove1 &&
	     ((mode = `Force) || (
		match checksum with
		    Some cs ->
		      let real_sum =
			try Digest.to_hex (Digest.file fullname)
			with
			    Sys_error msg ->
			      (* Cannot get digest value: Soft error, too *)
			      success := false;
			      message ("Error: " ^ msg ^ "\n"); ""
		      in
		      (cs = "MD5:" ^ real_sum) || (
			(* Checksum wrong: soft error, too *)
			success := false;
			message ("Checksum mismatch: " ^ fullname ^ "\n");
			if not !csm_explained then
			  message ("  (This means the file was modified. It is not deleted.)\n");
			csm_explained := true;
			false
		      )
		  | None -> true
	      )) in
	   if can_remove2 then (
	     match mode with
		 `Force | `Normal ->
		   ( try
		       (* Add directory to the set of directories to check
			* for removal:
			*)
		       let dir = Filename.dirname fullname in
		       dirs := StringSet.add dir !dirs;

		       (* Remove file: *)
		       Sys.remove fullname;
		       (* Maybe there is a backup file that must be restored: *)
		       if !preserve then (
			 let n = ipkg # name in
			 let backup_name = create_backup_name n fullname in
			 if Sys.file_exists backup_name then
			   Sys.rename backup_name fullname
		       );
		     with
			 Sys_error msg ->
			   (* Cannot delete/restore file: Soft error *)
			   success := false;
			   message ("Error: " ^ msg ^ "\n")
		   )
	       | `Simulate ->
		   ()
	   )

       | `Unexec(cmd) ->
	   check_prefix();
	   let cmd' = rewrite_placeholders !prefix !last_file cmd in
	   if mode = `Force || mode = `Normal then
	     ( try
		 exec_to_user_error
		   ~pref_user_error:true
		   (Godi_exec.call_cmd
		      ~stdout:protocol
		      ~stderr:protocol
		      ?interactivity)
		   (Godi_exec.shell cmd')
	       with
		   User_error msg ->
		     (* Cannot run script, or exit code <> 0: This is also
		      * a soft error.
		      *)
		     success := false;
		     message ("Error: " ^ msg ^ "\n")
	     )

       | _ ->
	   (* Ignore other directives! *)
	   ()
    )
    plist;

  (* Check which directories are now empty and can be removed: *)
  if mode <> `Simulate then (
    let l = ref [] in
    StringSet.iter (fun d -> l := d :: !l) !dirs;
    (* The directories in !l are now sorted in reverse order! This is
     * exactly what we need, because the deepest directories come first.
     *)
    List.iter
      (fun dir ->
	 let d = ref dir in
	 try
	   while !d <> "/" && !d <> "" && !d <> localbase do
	     Godi_file.remove_dir !d;  (* fails if non-empty *)
	     d := Filename.dirname !d;
	   done
	 with
	     Sys_error _ ->
	       (* Errors are ignored here. We will encounter them anyway,
		* because after some loop iterations the directory is
		* no longer empty, and [remove_dir] fails.
		*)
	       ()
      )
      !l
  );
  !success
;;


let remove_pkg_db_entry ?(mode = `Normal)
                        ?(protocol = Lazy.force default_protocol) pkg =
  (* Removes the db entries for pkg. The +REQUIRED_BY files of the predecessors
   * are updated, too.
   *
   * Returns [true] if no errors occurred, [false] if there were soft
   * errors. Hard errors trigger exceptions.
   *)
  let message = Godi_exec.collect protocol in
  let ipkg = installed_pkg pkg in
  let dir = ipkg # db_directory in

  match mode with
      `Normal | `Force ->
	let success = ref true in
	(* First delete the db directory: *)
	( try
	    Godi_file.remove_dir ~recurse:true dir
	  with
	      Sys_error msg ->
		(* If we cannot delete the db dir, this is considered
		 * as a hard error!
		 *)
		user_error("While deleting " ^ dir ^ ": " ^ msg)
	);
	(* Now update the "+REQUIRED_BY" files of the predecessor packages.
	 * These files are for informational purposes only, so failures
	 * are regarded as soft errors.
	 *)
	List.iter
	  (fun (dep_name, _) ->
	     try
	       let dpkg =
		 List.find                       (* may raise Not_found *)
		   (fun p -> pkg_name p = dep_name)
		   (get_installed_packages()) in
	       let filter n = n <> ipkg#name in
	       (installed_pkg dpkg) # update_successor_file ~filter ()
	     with
		 Not_found ->
		   (* A predecessor is missing ... - just ignore, we delete anyway *)
		   ()
	       | Sys_error msg ->
		   message("Error: " ^ msg ^ "\n");
		   success := false;
		   (* ... and continue *)
	  )
	  ipkg # dependencies;

	notify_about_change_of_installed_packages `Reconcile;
	!success

    | `Simulate ->
	true
;;


let remove_pkg ?(mode = `Normal) ?(soft_errors=false)
               ?interactivity ?(protocol = Lazy.force default_protocol) pkg =
  (* Deletes [pkg]. It is not checked whether there are still successor
   * packages (the planner will get this right). All scripts are run.
   *
   * It is distinguished between soft and hard errors. On soft errors,
   * the deletion continues, whereas hard errors stop the process.
   * Finally, a [User_error] exception is raised (in both cases).
   *)

  let message = Godi_exec.collect protocol in

  if pkgtype pkg <> `Installed then
    failwith "Godi_perform.remove_pkg: not an installed package";
  (* This is a programmatical error. *)

  let ipkg = installed_pkg pkg in
  let dir = ipkg # db_directory in
  let localbase = Godi_query.localbase() in

  (* If the db directory does not exist, assume the package is already
   * removed.
   *)
  let cont = Sys.file_exists dir || (
    message ("Warning: Package " ^ ipkg#name ^ " is already removed, skipping.\n");
    false
  ) in

  (* Maybe the package is protected *)
  (* Note: This is a warning only, because the overall algorithm should
   * usually continue.
   *)
  let cont =
    cont && (
      not(ipkg#preserve) || (
	message ("Warning: Package " ^ ipkg#name ^ " is protected against deletion.\n");
	(mode = `Force) && (
	  message "Deleting anyway.\n";
	  true)
      )) in

  (* Check whether there is a requirements script *)
  (* Note: This is a warning only, because the overall algorithm should
   * usually continue.
   *)
  let cont = cont && (
    match ipkg # requirements_script_path with
	None -> true
      | Some p ->
	  (* CHECK: chmod necessary? *)
	  let dir = Filename.dirname p in
	  let base = Filename.basename p in
	  try
	    exec_to_user_error
	      ~pref_user_error:true
	      (Godi_exec.call_cmd
		 ~cd:dir
		 ~env:["GODI_PREFIX", localbase]
		 ~stdout:protocol
		 ~stderr:protocol
		 ?interactivity)
	      (Godi_exec.shell
		 (sprintf "./%s '%s' DEINSTALL" base (pkg_fullname pkg)));
	    true
	  with
	      User_error m ->
		message("Warning: Requirements for deletion not given: " ^ m ^ "\n");
		(mode = `Force) && (
		  message("Continuing deletion anyway.\n");
		  true
		)
  ) in

  (* Perform the deletion: *)
  if cont then (
    (* Execute deinstall script: *)
    ( match ipkg # deinstall_script_path with
	  None -> ()
	| Some p ->
	    (* CHECK: chmod necessary? *)
	    let dir = Filename.dirname p in
	    let base = Filename.basename p in
	    ( try
		exec_to_user_error
		  ~pref_user_error:true
		  (Godi_exec.call_cmd
		     ~cd:dir
		     ~env:["GODI_PREFIX", localbase]
		     ~stdout:protocol
		     ~stderr:protocol
		     ?interactivity)
		  (Godi_exec.shell
		     (sprintf "./%s '%s' DEINSTALL" base (pkg_fullname pkg)));
	      with
		  User_error m ->
		    message("Error: Pre-deinstallation script fails: " ^ m ^ "\n");
		    if mode = `Force then
		      message("Continuing deletion anyway.\n")
		    else
		      user_error("Deletion of package " ^
				 pkg_fullname pkg ^ " failed")

	    )
    );

    (* Remove files, and db entry: *)
    let e1 = remove_pkg_files ~mode ?interactivity ~protocol pkg in
    let e2 = remove_pkg_db_entry ~mode ~protocol pkg in

    (* Execute deinstall script again: *)
    ( match ipkg # deinstall_script_path with
	  None -> ()
	| Some p ->
	    (* CHECK: chmod necessary? *)
	    let dir = Filename.dirname p in
	    let base = Filename.basename p in
	    ( try
		exec_to_user_error
		  ~pref_user_error:true
		  (Godi_exec.call_cmd
		     ~cd:dir
		     ~env:["GODI_PREFIX", localbase]
		     ~stdout:protocol
		     ~stderr:protocol
		     ?interactivity)
		  (Godi_exec.shell
		     (sprintf "./%s '%s' POST-DEINSTALL" base (pkg_fullname pkg)));
	      with
		  User_error m ->
		    message("Error: Post-deinstallation script fails: " ^ m ^ "\n");
		    if mode = `Force then
		      message("Continuing deletion anyway.\n")
		    else
		      user_error("Deletion of package " ^
				 pkg_fullname pkg ^ " failed")
	    )
    );

    if not e1 || not e2 then
      if soft_errors then
	raise Soft_rm_error
      else
	user_error("Deletion of package " ^ pkg_fullname pkg ^ " failed")
  )
;;


let extract_binary_archive ?interactivity
                           ?(protocol = Lazy.force default_protocol)
			   bin_arch =
  (* Extract the archive file, and put the contents into a temporary
   * directory. Returns the handle of this directory.
   *)
  let bin_stage = Godi_file.allocate_tmpdir() in
  exec_to_user_error
    ~pref_user_error:true
    (Godi_exec.call_cmd
       ~stdout:protocol
       ~stderr:protocol
       ~cd:(Godi_file.tmpdir_path bin_stage)
       ?interactivity)
    ("godi_tar", [| "xzef"; bin_arch |]);
  bin_stage
;;


let delete_installed_files prefix file =
  (* Delete the files whose paths are contained in the [file]. Ignore
   * problems.
   *)
  let s = read_file file in
  let lines = Str.split (Str.regexp "\n") s in
  List.iter
    (fun f ->
       let abs_f = Filename.concat prefix f in
       try Sys.remove abs_f with _ -> ()
    )
    lines
;;



let dotdot_re =
  Str.regexp "\\(^\\.\\./\\)\\|\\(.*/\\.\\.$\\)\\|\\(.*/\\.\\./\\)\\|\\(^\\.\\.$\\)" ;;

let security_check name =
  (* Check whether [name] can be securely used:
   * - It is a relative name
   * - It does not contain ".."
   *)
  Filename.is_relative name &&
  not(Str.string_match dotdot_re name 0)
;;


let get_binary_plist bin_stage =
  let plist_file = 
    Filename.concat (Godi_file.tmpdir_path bin_stage) "+CONTENTS" in
  let plist =
    Godi_plist.apply_ignore_directive
      (Godi_plist.parse_inst_plist plist_file) in
  plist
;;


type add_mode = [ `Normal | `Simulate ];;

let install_binary_archive ?(mode = `Normal)
                           ?interactivity
                           ?(protocol = Lazy.force default_protocol)
			   bin_stage plist =
  (* Moves the files from the staging area to their final place.
   * This is done by calling godi_pax with a suitable input.
   *)
  let message = Godi_exec.collect protocol in
  let localbase = Godi_query.localbase() in

  (* Get the plist information: *)
  let pname_full =
    try Godi_plist.extract_package_name plist
    with Not_found ->
      user_error("Error in packing list - missing @name directive") in
  let (pname, pname_version, pname_rev) =
    Godi_versions.split_name_version pname_full in

  (* Error handling: In [onerror], collect functions that clean things up
   * and undo actions.
   *)
  let onerror = ref [] in
  try

    (* Create the list files collecting the file names to move to the final
     * place: 
     * Because there may be several `Cwd directives in plist, we provide
     * one file per prefix. file_list_map maps prefixes to pairs
     * (fd, name) where [fd] is the file descriptor, and [name] is the
     * file name of the list.
     *)
    let file_list_stage = Godi_file.allocate_tmpdir() in
    let file_list_path = Godi_file.tmpdir_path file_list_stage in
    onerror := (fun () ->
		  Godi_file.release_tmpdir file_list_stage) :: !onerror;
    let file_list_map = ref StrMap.empty in
    let file_list_cnt = ref 0 in

    let append_to_list prefix line =
      try
	let (f,_) = StrMap.find prefix !file_list_map in
	output_string f (line ^ "\n")
      with
	  Not_found ->
	    let n = !file_list_cnt in
	    incr file_list_cnt;
	    let file = Filename.concat file_list_path (string_of_int n) in
	    let f = open_out_bin file in
	    output_string f (line ^ "\n");
	    file_list_map := StrMap.add prefix (f,file) !file_list_map
    in
    let close_lists() =
      StrMap.iter (fun _ (f,_) -> close_out f) !file_list_map;
    in
    onerror := close_lists :: !onerror;

    (* Further variables for PLIST interpretation: *)
    let prefix = ref "" in
    let last_file = ref "" in
    let preserve = ref false in

    let check_prefix() =
      (* A missing or bad prefix is a fatal error, because we cannot continue!
       *)
      (* Missing prefix: *)
      if !prefix = "" then
	user_error "Error in packing list - no prefix set";
      (* Prefix must be a subdirectory of [LOCALBASE] *)
      if not (localbase = !prefix || is_prefix_of (localbase ^ "/") !prefix) then
	user_error "Error in packing list - prefix outside of managed directories";
    in

    (* Iterate over the plist in a first pass, and check the integrity: *)
    List.iter
      (function
	 | `File(name,checksum) ->
	     (* Reject absolute file names, and occurrences of "..": *)
	     if not (security_check name) then
	       user_error("Error in packing list - insecure file name");
	     (* Basic sanity checks: *)
	     let stage_name = Filename.concat 
				(Godi_file.tmpdir_path bin_stage) name in
	     if not(Godi_file.file_exists ~symlink:true stage_name) then
	       user_error("Error in packing list - file does not exist: " ^ name);
	     if not(Godi_file.file_type ~is_symlink:true stage_name) &&
	       (Godi_file.file_type ~is_dir:true stage_name) then
		 user_error("Error in packing list - file is a directory: " ^ name);
	     (* Is the checksum correct? *)
	     ( match checksum with
		   None -> ()
		 | Some cs ->
		     if Godi_file.file_type ~is_symlink:true stage_name then
		       user_error("Error in packing list - checksum for symlink found: " ^ name);
		     let real_sum =
		       try Digest.to_hex (Digest.file stage_name)
		       with
			   Sys_error msg ->
			     user_error("While checking packing list: Cannot determine checksum: " ^ msg) in
		     if cs <> "MD5:" ^ real_sum then
		       user_error("Error in packing list - checksum mismatch: " ^ name)
	     );
	 | _ ->
	     ()
      )
      plist;

    (* Now iterate over the plist, and move the files/record the move: *)
    let cmds = ref [] in
    List.iter
      (function
	 | `Cwd d ->
	     if d <> "." then prefix := d
	     (* `Cwd "." is added by godi_create just before the "+" files *)
	 | `Option `Preserve ->
	     preserve := true
	     (* Enables the check for preserved files *)
	 | `File(name,checksum) ->
	     check_prefix();
	     last_file := name;
	     let fullname = Filename.concat !prefix name in
	     (* Check whether the file already exists, and which type the
	      * file has.
	      *)
	     if Godi_file.file_exists ~symlink:true fullname then (
	       let is_symlink = Godi_file.file_type ~is_symlink:true fullname in
	       let is_regular = Godi_file.file_type ~is_regular:true fullname in
	       if is_symlink || is_regular then (
		 (* Ok, we can handle this. *)
		 (* Maybe we have to preserve the old file: *)
		 if !preserve then (
		   let backup_name = create_backup_name pname fullname in
		   try
		     Sys.rename fullname backup_name;
		     onerror := (fun () -> Sys.rename backup_name fullname) ::
		       !onerror
		   with
		       Sys_error msg ->
			 user_error("Cannot move existing file away: " ^ msg)

		 )
		 else (
		   if is_symlink then
		     message("Replacing existing symlink " ^ fullname ^ "\n")
		     (* Actually, this is done by godi_pax later *)
		   else (
		     message("Replacing existing file " ^ fullname ^ "\n");
		     message("  (Keeping old version in <name>.godisav)\n");
		     Godi_file.copy ~prefer_link:true fullname (fullname ^ ".godisav")
		       (* Don't delete <name>.godisav in case of an error,
			* because <name> is going to be deleted then.
			*)
		   )
		 )
	       )
	       else
		 user_error("Cannot install a file over an existing directory: " ^ name)
	     );
	     (* Add the file to the right list: *)
	     append_to_list !prefix name;

	     (* CAVEAT: 
	      * We do not check whether parent directories of [name] are
	      * actually symlinks. This would be illegal, however.
	      *)

	 | `Exec cmd ->
	     (* Just record the command *)
	     let cmd' = rewrite_placeholders !prefix !last_file cmd in
	     cmds := cmd' :: !cmds

	 | `Unsupported(d,_) ->
	     message("Ignoring unsupported directive in packing list: " ^ d ^
		     "\n")
	 | _ ->
	     ()
      )
      plist;
    close_lists();
    check_prefix();

    (* By invoking godi_pax here, the move is finally done.
     * From here on, it is no longer possible to roll the operation fully back
     * when an error occurs:
     *
     * - As godi_pax may replace files it is not possible to undo this.
     *   We just ignore this problem!
     * - The @exec commands may do anything, we cannot control this, and
     *   it is unclear how to roll them back. In general, @unexec may be
     *   wrong (and we do not know which @unexec commands are the right
     *   ones).
     *   We also ignore this problem.
     *)
    ( match mode with
	  `Normal ->
	    onerror := (fun () -> 
			  StrMap.iter
  			  (fun list_prefix (_,list_name) ->
			     delete_installed_files list_prefix list_name)
			  !file_list_map) :: !onerror;
	    StrMap.iter
	      (fun list_prefix (_,list_name) ->
		 exec_to_user_error
		 ~pref_user_error:true
		 (Godi_exec.call_cmd
		    ~stdin:(`File list_name)
		    ~stdout:protocol
		    ~stderr:protocol
		    ~cd:(Godi_file.tmpdir_path bin_stage)
		    ?interactivity)
		 ("godi_pax", [| "-rw"; "-l"; "-p"; "p"; list_prefix |] )
	      )
	      !file_list_map;

	| `Simulate ->
	    ()
    );

    Godi_file.release_tmpdir file_list_stage;

    (* Invoke the @exec commands: *)
    ( match mode with
	  `Normal ->
	    List.iter
	    (fun cmd ->
	       exec_to_user_error
	       ~pref_user_error:true
	       (Godi_exec.call_cmd
		  ~stdout:protocol
		  ~stderr:protocol
		  ~cd:(Godi_file.tmpdir_path bin_stage)
		  ?interactivity)
	       (Godi_exec.shell cmd)
	    )
	    (List.rev !cmds)

	| `Simulate -> ()
    )
  with
      error ->
	List.iter (fun f -> f()) !onerror;
	raise error
;;


let add_pkg_db_entry ?(mode = `Normal)
                     ?(protocol = Lazy.force default_protocol)
                     bin_stage plist fullname
  =
  (* Create the directory, and copy the required files from bin_stage into
   * it. Write plist (it is modified).
   *)
  match mode with
      `Normal ->
	let dbdir = pkgdbdir() in
	let pkg_dbdir = Filename.concat dbdir fullname in
	Godi_file.create_dir pkg_dbdir;
	
	( try
	    let stage_dir = Godi_file.tmpdir_path bin_stage in
	    
	    let files =
	      List.filter
		(fun name -> name <> "" && name <> "+CONTENTS" && name.[0] = '+')
		(Godi_file.read_dir stage_dir) in
	    List.iter
	      (fun file ->
		 let contents = read_file ~pref_user_error:true
				  (Filename.concat stage_dir file) in
		 let f = open_out_bin (Filename.concat pkg_dbdir file) in
		 try
		   output_string f contents;
		   close_out f
		 with
		     error -> close_out f; raise error
	      )
	      files;
	    
	    Godi_plist.print_inst_plist (Filename.concat pkg_dbdir "+CONTENTS") plist
	  with
	      error ->
		(* Get rid of package dir again *)
		Godi_file.remove_dir ~recurse:true pkg_dbdir;
		raise error
	)

    | `Simulate ->
	()
;;


let add_pkg ?(mode = `Normal)
            ?interactivity ?(protocol = Lazy.force default_protocol)
            pkg =
  (* - Check whether package already installed. If so, error.
   * - Call extract_binary_archive, get bin_stage
   * - Execute requirements file (if any)
   * - Execute pre-install script (if any)
   * - Install the package: Call install_binary_archive
   * - Execute post-install script
   * - Update package db: add_pkg_db_entry
   * - Output display file, if any
   *)
  let message = Godi_exec.collect protocol in
  let localbase = Godi_query.localbase() in

  let dbdir = pkgdbdir() in
  let fullname = pkg_fullname pkg in
  let pkg_dbdir = Filename.concat dbdir fullname in
  if Sys.file_exists pkg_dbdir then
    user_error ("Package seems to be already installed: " ^ fullname);

  (* Error handling: *)
  let onerror = ref [] in
  try
    let bin_pkg = binary_pkg pkg in
    bin_pkg # require_binary_archive ~protocol ?interactivity ();
    let bin_arch = bin_pkg # binary_archive in
    onerror := (fun () -> bin_pkg # release_binary_archive()) :: !onerror;

    let bin_stage = extract_binary_archive ?interactivity ~protocol bin_arch in
    onerror := (fun () -> Godi_file.release_tmpdir bin_stage) :: !onerror;

    let plist = get_binary_plist bin_stage in

    (* Check whether there is the requirements file: *)
    let stage_dir = Godi_file.tmpdir_path bin_stage in
    let req_path = Filename.concat stage_dir "+REQUIRE" in
    if Sys.file_exists req_path then (
      (* CHECK: chmod necessary? *)
      try
	exec_to_user_error
	  ~pref_user_error:true
	  (Godi_exec.call_cmd
	     ~cd:stage_dir
	     ~env:["GODI_PREFIX", localbase]
	     ~stdout:protocol
	     ~stderr:protocol
	     ?interactivity)
	  (Godi_exec.shell
	     (sprintf "./+REQUIRE '%s' INSTALL" (pkg_fullname pkg)));
      with
	  User_error m ->
	    user_error("Requirements for addition not given: " ^ m);
    );

    (* Check whether there is an install script: *)
    let inst_path = Filename.concat stage_dir "+INSTALL" in
    if Sys.file_exists inst_path && mode = `Normal then (
      (* CHECK: chmod necessary? *)
      try
	exec_to_user_error
	  ~pref_user_error:true
	  (Godi_exec.call_cmd
	     ~cd:stage_dir
	     ~env:["GODI_PREFIX", localbase]
	     ~stdout:protocol
	     ~stderr:protocol
	     ?interactivity)
	  (Godi_exec.shell
	     (sprintf "./+INSTALL %s PRE-INSTALL" 
		(Filename.quote(pkg_fullname pkg))));
      with
	  User_error m ->
	    user_error("Pre-install script returned error status: " ^ m);
    );

    (* Update the package db: *)
    add_pkg_db_entry ~mode ~protocol bin_stage plist fullname;
    if mode = `Normal then
      onerror := (fun () -> Godi_file.remove_dir ~recurse:true pkg_dbdir) ::
        !onerror;

    (* Now install the files: *)
    install_binary_archive ~mode ~protocol ?interactivity bin_stage plist;

    (* From now on, there is not any error strategy. The package is installed,
     * only a few post-install actions are missing (that hardly can go
     * wrong).
     *)
    onerror := [];

    (* Execute post-install script, if any: *)
    if Sys.file_exists inst_path && mode = `Normal then (
      (* CHECK: chmod necessary? *)
      try
	exec_to_user_error
	  ~pref_user_error:true
	  (Godi_exec.call_cmd
	     ~cd:stage_dir
	     ~env:["GODI_PREFIX", localbase]
	     ~stdout:protocol
	     ~stderr:protocol
	     ?interactivity)
	  (Godi_exec.shell
	     (sprintf "./+INSTALL '%s' POST-INSTALL" (pkg_fullname pkg)));
      with
	  User_error m ->
	    message("Warning: Post-install script returned error status: " ^ m ^ "\n");
	    message("Warning: The package " ^ fullname ^ " was not cleanly installed, please check manually\n");
    );

    (* Output display file, if any: *)
    ( try
	let disp_el = 
	  List.find (function `Display f -> true | _ -> false) plist in
	(* or Not_found *)
	let disp_file =
	  match disp_el with `Display f -> f | _ -> assert false in
	let disp_full_file = Filename.concat stage_dir disp_file in
	if not (Sys.file_exists disp_full_file) then
	  message("Warning: Messgage file " ^ disp_file ^ " not found\n")
	else (
	  message(read_file ~pref_user_error:true disp_full_file)
	)
      with
	  Not_found -> (* No display file *)
	    ()
	| User_error msg ->
	    message("Warning: While reading message file: " ^ msg);
    );

    (* Cleanup: *)
    Godi_file.release_tmpdir bin_stage;
    bin_pkg # release_binary_archive();
    onerror := [];

    (* Maybe we have to update the +REQUIRED_BY file *)
    if mode = `Normal then (
      notify_about_change_of_installed_packages `Reconcile;

      let inst_pkg =
	try
	  List.find
	    (fun p -> pkg_name p = pkg_name pkg)
	    (get_installed_packages ~protocol ())
	with
	    Not_found ->
	      failwith "Cannot find the package that was just installed!" in
      (installed_pkg inst_pkg) # update_successor_file ()
    )

  with
      error ->
	List.iter (fun e -> e()) !onerror;
	raise error
;;


(**********************************************************************)
(* Source packages                                                    *)
(**********************************************************************)

let fetch_pkg ?(pref_user_error=false) ?interactivity
              ?(protocol = Lazy.force default_protocol) pkg =
  if pkgtype pkg <> `Source then
    failwith "Godi_perform.fetch_pkg: not a source package";
  let dir = (source_pkg pkg) # source_directory in
  exec_to_user_error
    ~pref_user_error
    (Godi_exec.call_cmd
       ~cd:dir
       ~stdout:protocol
       ~stderr:protocol
       ?interactivity)
    ("godi_make", [| "fetch" |] )
;;


let test_conf_pkg ?(pref_user_error=false) ?interactivity
                  ?(protocol = Lazy.force default_protocol)
                  ?(godi_conf = Godi_query.godi_conf()) pkg =
  if pkgtype pkg <> `Source then
    failwith "Godi_perform.test_conf_pkg: not a source package";
  let message = Godi_exec.collect protocol in
  let dir = (source_pkg pkg) # source_directory in
  exec_to_user_error
    ~pref_user_error
    (fun () ->
       Godi_exec.call_cmd
       ~cd:dir
       ~stdout:protocol
       ~stderr:protocol
       ?interactivity
       ("godi_make", [| "clean" |] );
       (* Classic source package *)
       Godi_exec.call_cmd
	 ~cd:dir
	 ~env:[ "GODI_CONF", godi_conf ]
	 ~stdout:protocol
	 ~stderr:protocol
	 ?interactivity
	 ("godi_make", [| "configure" |] );
       Godi_exec.call_cmd
	 ~cd:dir
	 ~stdout:protocol
	 ~stderr:protocol
	 ?interactivity
	 ("godi_make", [| "clean" |] )
    )
    ()
;;


let build_pkg ?(pref_user_error=false) ?interactivity
              ?(protocol = Lazy.force default_protocol)
              ?(catch_self = Some "godi-tools") pkg =
  if pkgtype pkg <> `Source then
    failwith "Godi_perform.build_pkg: not a source package";
  let message = Godi_exec.collect protocol in
  let dir = (source_pkg pkg) # source_directory in
  exec_to_user_error
    ~pref_user_error
    (fun () ->
       (* TODO:
	* - catch_self is only applied when the package is already installed.
	*   Currently godi_delete outputs an error, but the error is ignored.
	*)
       Godi_exec.call_cmd
         ~cd:dir
         ~stdout:protocol
         ~stderr:protocol
         ?interactivity
         ("godi_make", [| "clean" |] );
       if (source_pkg pkg)#self_update_pkg <> "" then (
	 (* Package to self-update GODI: *)
	 Godi_exec.call_cmd
	   ~cd:dir
	   ~stdout:protocol
	   ~stderr:protocol
	   ?interactivity
	   ("godi_make", [| "self_package" |] );
	 (* Sleep here for a second, to work around a bug in godi_tar.
	  * This little pause can go away when it is unlikely that old
	  * versions of godi_tar are around.
	  *)
	 Unix.sleep 1;
	 (* If this package is also meant by catch_self, we do not replace
	  * the package immediately. Instead, we exit the program now and
	  * execute shell commands replacing the package followed by a
	  * restart of the program.
	  *)
	 if (catch_self = Some (source_pkg pkg)#name) then (
	   let tmpdir = Godi_file.create_tmp_dir() in
	   let xfilename = Filename.basename Sys.executable_name in
	   let self_copy = Filename.concat tmpdir xfilename in
	   (* TODO: cp, rm are called by implicit path *)
	   let cmd0 = sprintf "cp %s %s\n" Sys.executable_name self_copy in
	   let cmd1 = sprintf "%s delete -f %s\n" 
			self_copy (Filename.quote (source_pkg pkg)#name) in
	   let cmd2 = sprintf "%s add %s\n"
			self_copy
			(Filename.quote (source_pkg pkg)#self_update_pkg) in
	   let cmd3 = sprintf "rm -rf %s\n" tmpdir in
	   let sh_cmd = cmd0 ^ cmd1 ^ cmd2 ^ cmd3 in
	   raise (Restart_recommended sh_cmd)
	 );
	 (* In the normal case, run these commands directly: *)
	 (* Force deletion, ignore errors: *)
	 ( try
	     let old_pkg = 
	       try
		 get_this_installed_package ~protocol (pkg_name pkg)
	       with
		   Not_found -> 
		     message ("Warning: Package was not installed: " ^
			      (pkg_name pkg) ^ "\n");
		     raise Not_found in
	     remove_pkg ~mode:`Force ?interactivity ~protocol old_pkg;
	   with
	       User_error m ->
		 (* Ignore errors; just print them. *)
		 message("Warning: " ^ m ^ "\n");
	     | Not_found ->
		 ()
	 );
	 let new_pkg = get_binary_package (source_pkg pkg)#self_update_pkg in
	 add_pkg ?interactivity ~protocol new_pkg;
       ) else (
	 (* Classic source package *)
	 if (catch_self = Some (source_pkg pkg)#name) then (
	   (* We cannot handle this case. Output a cryptic warning: *)
	   Godi_exec.collect protocol
	     "Warning: This package may fail to replace itself\n";
	 );
	 Godi_exec.call_cmd
		 ~cd:dir
		 ~stdout:protocol
		 ~stderr:protocol
		 ?interactivity
		 ("godi_make", [| "package" |] );
       );
       Godi_exec.call_cmd
	 ~cd:dir
	 ~stdout:protocol
	 ~stderr:protocol
	 ?interactivity
	 ("godi_make", [| "clean" |] );
    )
    ()
;;

